# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .util import FTL_ATTR_SHIFT_HUMAN
from .util import generate_spec_record
from .util import generate_spec_attributes

# Generic BGP record type attribute
FTL_ATTR_BGP_RECORD_TYPE = 1 << 0

# Generic BGP record type attribute (human-readable)
FTL_ATTR_BGP_RECORD_TYPE_HUMAN = FTL_ATTR_BGP_RECORD_TYPE << FTL_ATTR_SHIFT_HUMAN

#############################
# BGP PEER_TABLE ATTRIBUTES #
#############################

# Available BGP peer_table attributes
FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID = 1 << 1
FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME        = 1 << 2
FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL    = 1 << 3
FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID      = 1 << 4
FTL_ATTR_BGP_PEER_TABLE_PEER_AS          = 1 << 5
FTL_ATTR_BGP_PEER_TABLE_PEER_IP          = 1 << 6

# Available BGP peer_table attributes (human-readable)
FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID_HUMAN = FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME_HUMAN        = FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL_HUMAN    = FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID_HUMAN      = FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_PEER_TABLE_PEER_AS_HUMAN          = FTL_ATTR_BGP_PEER_TABLE_PEER_AS          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_PEER_TABLE_PEER_IP_HUMAN          = FTL_ATTR_BGP_PEER_TABLE_PEER_IP          << FTL_ATTR_SHIFT_HUMAN

# BGP peer_table attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpPeerTable = generate_spec_record('FtlAttrsBgpPeerTable', (
    ('collector_bgp_id', FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID, False),
    ('view_name',        FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME,        False),
    ('peer_protocol',    FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL,    True),
    ('peer_bgp_id',      FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID,      True),
    ('peer_as',          FTL_ATTR_BGP_PEER_TABLE_PEER_AS,          True),
    ('peer_ip',          FTL_ATTR_BGP_PEER_TABLE_PEER_IP,          True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

###############################
# BGP STATE_CHANGE ATTRIBUTES #
###############################

# Available BGP state_change attributes
FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP     = 1 << 1
FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL = 1 << 2
FTL_ATTR_BGP_STATE_CHANGE_PEER_AS       = 1 << 3
FTL_ATTR_BGP_STATE_CHANGE_PEER_IP       = 1 << 4
FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE     = 1 << 5
FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE     = 1 << 6

# Available BGP state_change attributes (human-readable)
FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP_HUMAN     = FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL_HUMAN = FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATE_CHANGE_PEER_AS_HUMAN       = FTL_ATTR_BGP_STATE_CHANGE_PEER_AS       << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATE_CHANGE_PEER_IP_HUMAN       = FTL_ATTR_BGP_STATE_CHANGE_PEER_IP       << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE_HUMAN     = FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE_HUMAN     = FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE     << FTL_ATTR_SHIFT_HUMAN

# BGP state_change attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpStateChange = generate_spec_record('FtlAttrsBgpStateChange', (
    ('timestamp',     FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP,     True),
    ('peer_protocol', FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL, True),
    ('peer_as',       FTL_ATTR_BGP_STATE_CHANGE_PEER_AS,       True),
    ('peer_ip',       FTL_ATTR_BGP_STATE_CHANGE_PEER_IP,       True),
    ('old_state',     FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE,     True),
    ('new_state',     FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE,     True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

########################
# BGP ROUTE ATTRIBUTES #
########################

# Available BGP route attributes
FTL_ATTR_BGP_ROUTE_SOURCE               = 1 << 1
FTL_ATTR_BGP_ROUTE_SEQUENCE             = 1 << 2
FTL_ATTR_BGP_ROUTE_TIMESTAMP            = 1 << 3
FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL        = 1 << 4
FTL_ATTR_BGP_ROUTE_PEER_BGP_ID          = 1 << 5
FTL_ATTR_BGP_ROUTE_PEER_AS              = 1 << 6
FTL_ATTR_BGP_ROUTE_PEER_IP              = 1 << 7
FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL     = 1 << 8
FTL_ATTR_BGP_ROUTE_NEXTHOP_IP           = 1 << 9
FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL      = 1 << 10
FTL_ATTR_BGP_ROUTE_PREFIX               = 1 << 11
FTL_ATTR_BGP_ROUTE_PATH_ID              = 1 << 12
FTL_ATTR_BGP_ROUTE_ASPATH               = 1 << 13
FTL_ATTR_BGP_ROUTE_ORIGIN               = 1 << 14
FTL_ATTR_BGP_ROUTE_COMMUNITIES          = 1 << 15
FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES    = 1 << 16
FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES = 1 << 17
FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC      = 1 << 18
FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE     = 1 << 19
FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL  = 1 << 20
FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS        = 1 << 21
FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP        = 1 << 22
FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER     = 1 << 23
FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID        = 1 << 24
FTL_ATTR_BGP_ROUTE_CLUSTER_LIST         = 1 << 25
FTL_ATTR_BGP_ROUTE_LOCAL_PREF           = 1 << 26
FTL_ATTR_BGP_ROUTE_ATTR_SET             = 1 << 27
FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT         = 1 << 28
FTL_ATTR_BGP_ROUTE_AIGP                 = 1 << 29
FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN        = 1 << 30

# Available BGP route attributes (human-readable)
FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN               = FTL_ATTR_BGP_ROUTE_SOURCE               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_SEQUENCE_HUMAN             = FTL_ATTR_BGP_ROUTE_SEQUENCE             << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN            = FTL_ATTR_BGP_ROUTE_TIMESTAMP            << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN        = FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PEER_BGP_ID_HUMAN          = FTL_ATTR_BGP_ROUTE_PEER_BGP_ID          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PEER_AS_HUMAN              = FTL_ATTR_BGP_ROUTE_PEER_AS              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN              = FTL_ATTR_BGP_ROUTE_PEER_IP              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN     = FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN           = FTL_ATTR_BGP_ROUTE_NEXTHOP_IP           << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN      = FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN               = FTL_ATTR_BGP_ROUTE_PREFIX               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_PATH_ID_HUMAN              = FTL_ATTR_BGP_ROUTE_PATH_ID              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ASPATH_HUMAN               = FTL_ATTR_BGP_ROUTE_ASPATH               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN               = FTL_ATTR_BGP_ROUTE_ORIGIN               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_COMMUNITIES_HUMAN          = FTL_ATTR_BGP_ROUTE_COMMUNITIES          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES_HUMAN    = FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES_HUMAN = FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC_HUMAN      = FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE_HUMAN     = FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL_HUMAN  = FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS_HUMAN        = FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP_HUMAN        = FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER_HUMAN     = FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID_HUMAN        = FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_CLUSTER_LIST_HUMAN         = FTL_ATTR_BGP_ROUTE_CLUSTER_LIST         << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_LOCAL_PREF_HUMAN           = FTL_ATTR_BGP_ROUTE_LOCAL_PREF           << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ATTR_SET_HUMAN             = FTL_ATTR_BGP_ROUTE_ATTR_SET             << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT_HUMAN         = FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT         << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_AIGP_HUMAN                 = FTL_ATTR_BGP_ROUTE_AIGP                 << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN_HUMAN        = FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN        << FTL_ATTR_SHIFT_HUMAN

# BGP route attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpRoute = generate_spec_record('FtlAttrsBgpRoute', (
    ('source',               FTL_ATTR_BGP_ROUTE_SOURCE,               True),
    ('sequence',             FTL_ATTR_BGP_ROUTE_SEQUENCE,             True),
    ('timestamp',            FTL_ATTR_BGP_ROUTE_TIMESTAMP,            True),
    ('peer_protocol',        FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL,        True),
    ('peer_bgp_id',          FTL_ATTR_BGP_ROUTE_PEER_BGP_ID,          True),
    ('peer_as',              FTL_ATTR_BGP_ROUTE_PEER_AS,              True),
    ('peer_ip',              FTL_ATTR_BGP_ROUTE_PEER_IP,              True),
    ('nexthop_protocol',     FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL,     True),
    ('nexthop_ip',           FTL_ATTR_BGP_ROUTE_NEXTHOP_IP,           True),
    ('prefix_protocol',      FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL,      True),
    ('prefix',               FTL_ATTR_BGP_ROUTE_PREFIX,               True),
    ('path_id',              FTL_ATTR_BGP_ROUTE_PATH_ID,              True),
    ('aspath',               FTL_ATTR_BGP_ROUTE_ASPATH,               True),
    ('origin',               FTL_ATTR_BGP_ROUTE_ORIGIN,               True),
    ('communities',          FTL_ATTR_BGP_ROUTE_COMMUNITIES,          True),
    ('large_communities',    FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES,    True),
    ('extended_communities', FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES, False),
    ('multi_exit_disc',      FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC,      False),
    ('atomic_aggregate',     FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE,     False),
    ('aggregator_protocol',  FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL,  False),
    ('aggregator_as',        FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS,        False),
    ('aggregator_ip',        FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP,        False),
    ('only_to_customer',     FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER,     False),
    ('originator_id',        FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID,        False),
    ('cluster_list',         FTL_ATTR_BGP_ROUTE_CLUSTER_LIST,         False),
    ('local_pref',           FTL_ATTR_BGP_ROUTE_LOCAL_PREF,           False),
    ('attr_set',             FTL_ATTR_BGP_ROUTE_ATTR_SET,             False),
    ('as_pathlimit',         FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT,         False),
    ('aigp',                 FTL_ATTR_BGP_ROUTE_AIGP,                 False),
    ('attrs_unknown',        FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN,        False),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

#############################
# BGP KEEP_ALIVE ATTRIBUTES #
#############################

# Available BGP keep_alive attributes
FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP     = 1 << 1
FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL = 1 << 2
FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS       = 1 << 3
FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP       = 1 << 4

# Available BGP keep_alive attributes (human-readable)
FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP_HUMAN     = FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL_HUMAN = FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS_HUMAN       = FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS       << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP_HUMAN       = FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP       << FTL_ATTR_SHIFT_HUMAN

# BGP keep_alive attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpKeepAlive = generate_spec_record('FtlAttrsBgpKeepAlive', (
    ('timestamp',     FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP,     True),
    ('peer_protocol', FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL, True),
    ('peer_as',       FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS,       True),
    ('peer_ip',       FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP,       True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

################################
# BGP ROUTE_REFRESH ATTRIBUTES #
################################

# Available BGP route_refresh attributes
FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP        = 1 << 1
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL    = 1 << 2
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS          = 1 << 3
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP          = 1 << 4
FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL = 1 << 5

# Available BGP route_refresh attributes (human-readable)
FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP_HUMAN        = FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL_HUMAN    = FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS_HUMAN          = FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP_HUMAN          = FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL_HUMAN = FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL << FTL_ATTR_SHIFT_HUMAN

# BGP route_refresh attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpRouteRefresh = generate_spec_record('FtlAttrsBgpRouteRefresh', (
    ('timestamp',        FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP,        True),
    ('peer_protocol',    FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL,    True),
    ('peer_as',          FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS,          True),
    ('peer_ip',          FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP,          True),
    ('refresh_protocol', FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL, True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

###############################
# BGP NOTIFICATION ATTRIBUTES #
###############################

# Available BGP notification attributes
FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP     = 1 << 1
FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL = 1 << 2
FTL_ATTR_BGP_NOTIFICATION_PEER_AS       = 1 << 3
FTL_ATTR_BGP_NOTIFICATION_PEER_IP       = 1 << 4
FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE    = 1 << 5
FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE = 1 << 6
FTL_ATTR_BGP_NOTIFICATION_DATA          = 1 << 7

# Available BGP notification attributes (human-readable)
FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP_HUMAN     = FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP     << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL_HUMAN = FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_PEER_AS_HUMAN       = FTL_ATTR_BGP_NOTIFICATION_PEER_AS       << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_PEER_IP_HUMAN       = FTL_ATTR_BGP_NOTIFICATION_PEER_IP       << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_HUMAN    = FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_HUMAN = FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_NOTIFICATION_DATA_HUMAN          = FTL_ATTR_BGP_NOTIFICATION_DATA          << FTL_ATTR_SHIFT_HUMAN

# BGP notification attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpNotification = generate_spec_record('FtlAttrsBgpNotification', (
    ('timestamp',     FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP,     True),
    ('peer_protocol', FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL, True),
    ('peer_as',       FTL_ATTR_BGP_NOTIFICATION_PEER_AS,       True),
    ('peer_ip',       FTL_ATTR_BGP_NOTIFICATION_PEER_IP,       True),
    ('error_code',    FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE,    True),
    ('error_subcode', FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE, True),
    ('data',          FTL_ATTR_BGP_NOTIFICATION_DATA,          True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

#######################
# BGP OPEN ATTRIBUTES #
#######################

# Available BGP open attributes
FTL_ATTR_BGP_OPEN_TIMESTAMP      = 1 << 1
FTL_ATTR_BGP_OPEN_PEER_PROTOCOL  = 1 << 2
FTL_ATTR_BGP_OPEN_PEER_AS        = 1 << 3
FTL_ATTR_BGP_OPEN_PEER_IP        = 1 << 4
FTL_ATTR_BGP_OPEN_VERSION        = 1 << 5
FTL_ATTR_BGP_OPEN_MY_AS          = 1 << 6
FTL_ATTR_BGP_OPEN_HOLD_TIME      = 1 << 7
FTL_ATTR_BGP_OPEN_BGP_ID         = 1 << 8
FTL_ATTR_BGP_OPEN_CAPABILITIES   = 1 << 9
FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN = 1 << 10

# Available BGP open attributes (human-readable)
FTL_ATTR_BGP_OPEN_TIMESTAMP_HUMAN      = FTL_ATTR_BGP_OPEN_TIMESTAMP      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_PEER_PROTOCOL_HUMAN  = FTL_ATTR_BGP_OPEN_PEER_PROTOCOL  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_PEER_AS_HUMAN        = FTL_ATTR_BGP_OPEN_PEER_AS        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_PEER_IP_HUMAN        = FTL_ATTR_BGP_OPEN_PEER_IP        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_VERSION_HUMAN        = FTL_ATTR_BGP_OPEN_VERSION        << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_MY_AS_HUMAN          = FTL_ATTR_BGP_OPEN_MY_AS          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_HOLD_TIME_HUMAN      = FTL_ATTR_BGP_OPEN_HOLD_TIME      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_BGP_ID_HUMAN         = FTL_ATTR_BGP_OPEN_BGP_ID         << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_CAPABILITIES_HUMAN   = FTL_ATTR_BGP_OPEN_CAPABILITIES   << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN_HUMAN = FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN << FTL_ATTR_SHIFT_HUMAN

# BGP open attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpOpen = generate_spec_record('FtlAttrsBgpOpen', (
    ('timestamp',      FTL_ATTR_BGP_OPEN_TIMESTAMP,      True),
    ('peer_protocol',  FTL_ATTR_BGP_OPEN_PEER_PROTOCOL,  True),
    ('peer_as',        FTL_ATTR_BGP_OPEN_PEER_AS,        True),
    ('peer_ip',        FTL_ATTR_BGP_OPEN_PEER_IP,        True),
    ('version',        FTL_ATTR_BGP_OPEN_VERSION,        True),
    ('my_as',          FTL_ATTR_BGP_OPEN_MY_AS,          True),
    ('hold_time',      FTL_ATTR_BGP_OPEN_HOLD_TIME,      True),
    ('bgp_id',         FTL_ATTR_BGP_OPEN_BGP_ID,         True),
    ('capabilities',   FTL_ATTR_BGP_OPEN_CAPABILITIES,   True),
    ('params_unknown', FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN, False),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

########################
# BGP STATS ATTRIBUTES #
########################

# Available BGP stats attributes
FTL_ATTR_BGP_STATS_PARSER_LIFETIME          = 1 << 1
FTL_ATTR_BGP_STATS_PARSER_RUNTIME           = 1 << 2
FTL_ATTR_BGP_STATS_PARSER_ERRORS            = 1 << 3
FTL_ATTR_BGP_STATS_LGL_RUNTIME              = 1 << 4
FTL_ATTR_BGP_STATS_LGL_ENTRIES              = 1 << 5
FTL_ATTR_BGP_STATS_LGL_ERRORS               = 1 << 6
FTL_ATTR_BGP_STATS_MRT_RUNTIME              = 1 << 7
FTL_ATTR_BGP_STATS_MRT_ENTRIES              = 1 << 8
FTL_ATTR_BGP_STATS_MRT_ERRORS               = 1 << 9
FTL_ATTR_BGP_STATS_MRT_FIXES                = 1 << 10
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES      = 1 << 11
FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES    = 1 << 12
FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES  = 1 << 13
FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES = 1 << 14
FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4      = 1 << 15
FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6      = 1 << 16
FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4 = 1 << 17
FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6 = 1 << 18
FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4 = 1 << 19
FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6 = 1 << 20

# Available BGP stats attributes (human-readable)
FTL_ATTR_BGP_STATS_PARSER_LIFETIME_HUMAN          = FTL_ATTR_BGP_STATS_PARSER_LIFETIME          << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_PARSER_RUNTIME_HUMAN           = FTL_ATTR_BGP_STATS_PARSER_RUNTIME           << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_PARSER_ERRORS_HUMAN            = FTL_ATTR_BGP_STATS_PARSER_ERRORS            << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_LGL_RUNTIME_HUMAN              = FTL_ATTR_BGP_STATS_LGL_RUNTIME              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_LGL_ENTRIES_HUMAN              = FTL_ATTR_BGP_STATS_LGL_ENTRIES              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_LGL_ERRORS_HUMAN               = FTL_ATTR_BGP_STATS_LGL_ERRORS               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_RUNTIME_HUMAN              = FTL_ATTR_BGP_STATS_MRT_RUNTIME              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_ENTRIES_HUMAN              = FTL_ATTR_BGP_STATS_MRT_ENTRIES              << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_ERRORS_HUMAN               = FTL_ATTR_BGP_STATS_MRT_ERRORS               << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN                = FTL_ATTR_BGP_STATS_MRT_FIXES                << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES_HUMAN      = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES_HUMAN    = FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES_HUMAN  = FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES_HUMAN = FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4_HUMAN      = FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6_HUMAN      = FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6      << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4_HUMAN = FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4 << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6_HUMAN = FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6 << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4_HUMAN = FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4 << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6_HUMAN = FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6 << FTL_ATTR_SHIFT_HUMAN

# BGP stats attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpStats = generate_spec_record('FtlAttrsBgpStats', (
    ('parser_lifetime',          FTL_ATTR_BGP_STATS_PARSER_LIFETIME,          True),
    ('parser_runtime',           FTL_ATTR_BGP_STATS_PARSER_RUNTIME,           True),
    ('parser_errors',            FTL_ATTR_BGP_STATS_PARSER_ERRORS,            True),
    ('lgl_runtime',              FTL_ATTR_BGP_STATS_LGL_RUNTIME,              True),
    ('lgl_entries',              FTL_ATTR_BGP_STATS_LGL_ENTRIES,              True),
    ('lgl_errors',               FTL_ATTR_BGP_STATS_LGL_ERRORS,               True),
    ('mrt_runtime',              FTL_ATTR_BGP_STATS_MRT_RUNTIME,              True),
    ('mrt_entries',              FTL_ATTR_BGP_STATS_MRT_ENTRIES,              True),
    ('mrt_errors',               FTL_ATTR_BGP_STATS_MRT_ERRORS,               True),
    ('mrt_fixes',                FTL_ATTR_BGP_STATS_MRT_FIXES,                True),
    ('mrt_bgp_entry_types',      FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES,      True),
    ('mrt_bgp_message_types',    FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES,    True),
    ('mrt_bgp_attribute_types',  FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES,  True),
    ('mrt_bgp_capability_types', FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES, True),
    ('bgp_routes_rib_ipv4',      FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4,      True),
    ('bgp_routes_rib_ipv6',      FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6,      True),
    ('bgp_routes_announce_ipv4', FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4, True),
    ('bgp_routes_announce_ipv6', FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6, True),
    ('bgp_routes_withdraw_ipv4', FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4, True),
    ('bgp_routes_withdraw_ipv6', FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6, True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

########################
# BGP ERROR ATTRIBUTES #
########################

# Available BGP error attributes
FTL_ATTR_BGP_ERROR_SOURCE  = 1 << 1
FTL_ATTR_BGP_ERROR_SCOPE   = 1 << 2
FTL_ATTR_BGP_ERROR_RECORD  = 1 << 3
FTL_ATTR_BGP_ERROR_REASON  = 1 << 4
FTL_ATTR_BGP_ERROR_MESSAGE = 1 << 5
FTL_ATTR_BGP_ERROR_DATA    = 1 << 6
FTL_ATTR_BGP_ERROR_TRACE   = 1 << 7

# Available BGP error attributes (human-readable)
FTL_ATTR_BGP_ERROR_SOURCE_HUMAN  = FTL_ATTR_BGP_ERROR_SOURCE  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_SCOPE_HUMAN   = FTL_ATTR_BGP_ERROR_SCOPE   << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_RECORD_HUMAN  = FTL_ATTR_BGP_ERROR_RECORD  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_REASON_HUMAN  = FTL_ATTR_BGP_ERROR_REASON  << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_MESSAGE_HUMAN = FTL_ATTR_BGP_ERROR_MESSAGE << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_DATA_HUMAN    = FTL_ATTR_BGP_ERROR_DATA    << FTL_ATTR_SHIFT_HUMAN
FTL_ATTR_BGP_ERROR_TRACE_HUMAN   = FTL_ATTR_BGP_ERROR_TRACE   << FTL_ATTR_SHIFT_HUMAN

# BGP error attribute specification
# [Format] spec := (field, value, default)
FtlAttrsBgpError = generate_spec_record('FtlAttrsBgpError', (
    ('source',  FTL_ATTR_BGP_ERROR_SOURCE,  True),
    ('scope',   FTL_ATTR_BGP_ERROR_SCOPE,   True),
    ('record',  FTL_ATTR_BGP_ERROR_RECORD,  True),
    ('reason',  FTL_ATTR_BGP_ERROR_REASON,  True),
    ('message', FTL_ATTR_BGP_ERROR_MESSAGE, True),
    ('data',    FTL_ATTR_BGP_ERROR_DATA,    True),
    ('trace',   FTL_ATTR_BGP_ERROR_TRACE,   True),
), typed=FTL_ATTR_BGP_RECORD_TYPE)

######################
# ALL BGP ATTRIBUTES #
######################

# BGP attribute spec
FtlAttrsBgp = generate_spec_attributes('FtlAttrsBgp', (
    FtlAttrsBgpPeerTable,
    FtlAttrsBgpStateChange,
    FtlAttrsBgpRoute,
    FtlAttrsBgpKeepAlive,
    FtlAttrsBgpRouteRefresh,
    FtlAttrsBgpNotification,
    FtlAttrsBgpOpen,
    FtlAttrsBgpStats,
    FtlAttrsBgpError,
))



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .entry.bgp import unpack_mrt_entry_bgp
from .entry.bgp4mp import unpack_mrt_entry_bgp4mp
from .entry.tdv2 import unpack_mrt_entry_tdv2_index
from .entry.tdv2 import unpack_mrt_entry_tdv2_rib
from .entry.td import unpack_mrt_entry_td_rib
from .entry.const import MRT_VALID
from .entry.const import MRT_BGP
from .entry.const import MRT_BGP4MP
from .entry.const import MRT_BGP4MP_ET
from .entry.const import MRT_TABLE_DUMP
from .entry.const import MRT_TABLE_DUMP_V2
from .entry.const import MRT_TABLE_DUMP_V2_PEER_INDEX_TABLE
from .entry.const import MRT_TABLE_DUMP_V2_RIB_ADDPATH_ANY
from .entry.const import MRT_HEADER_BYTES
from .entry.const import MRT_DATA_BYTES
from ..const import STRUCT_2B
from ..const import STRUCT_4B
from ..const import struct_unpack
from ...parser import FtlParserFunc
from ...model.attr import FTL_ATTR_BGP_STATS_MRT_ENTRIES
from ...model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES
from ...model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES_HUMAN
from ...model.const import FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPE_TO_STR
from ...model.const import FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TO_STR
from ...model.record import FTL_RECORD_BGP_STATS
from ...model.error import FtlMrtError
from ...model.error import FtlMrtFormatError
from ...model.error import FtlMrtHeaderError
from ...model.error import FtlMrtDataError


@FtlParserFunc
def unpack_mrt_data(inputfile, caches, stats_record, bgp_records, bgp_error):
    """ Parse MRT byte stream.
    """
    # Prepare peer table data structure and sequence number
    peer_table, sequence = list(), 0

    # Prepare memory for MRT header bytes (fixed length)
    header_bytes = memoryview(bytearray(MRT_HEADER_BYTES))

    # Prepare MRT header fields
    ts, mrt_type, mrt_subtype, length = None, None, None, None

    # Prepare memory for MRT data bytes (dynamic length)
    data_bytes = memoryview(bytearray(MRT_DATA_BYTES))
    entry_bytes = memoryview(bytearray())

    # Access BGP record templates
    peer_table_records = bgp_records.peer_table
    state_change_records = bgp_records.state_change
    route_records = bgp_records.route
    keep_alive_records = bgp_records.keep_alive
    route_refresh_records = bgp_records.route_refresh
    notification_records = bgp_records.notification
    open_records = bgp_records.open

    # Read MRT byte stream
    first_record = True
    while True:
        try:
            # Read MRT header bytes
            n_read = 0
            try:
                n_read = inputfile.readinto(header_bytes)
            except Exception:  # pylint: disable=broad-except
                pass

            # Check MRT header bytes
            if n_read < MRT_HEADER_BYTES:
                if n_read > 0:
                    raise FtlMrtHeaderError(f'Incomplete MRT header ({n_read}B<{MRT_HEADER_BYTES}B)')
                break

            # ------------------------------
            # [RFC6396] 2. MRT Common Header
            # ------------------------------
            #  0                   1                   2                   3
            #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            # |                           Timestamp                           |
            # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            # |             Type              |            Subtype            |
            # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            # |                             Length                            |
            # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            # |                      Message... (variable)
            # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            # Parse MRT header data
            ts = float(struct_unpack(STRUCT_4B, header_bytes[:4])[0])
            mrt_type = struct_unpack(STRUCT_2B, header_bytes[4:6])[0]
            mrt_subtype = struct_unpack(STRUCT_2B, header_bytes[6:8])[0]
            length = struct_unpack(STRUCT_4B, header_bytes[8:12])[0]

            # Check MRT entry type
            if mrt_type not in MRT_VALID:
                raise FtlMrtHeaderError(f'Unknown MRT type ({mrt_type})', data=header_bytes)

            # Check MRT entry length
            if length == 0:
                raise FtlMrtHeaderError('Empty MRT record', data=header_bytes)
            if length > MRT_DATA_BYTES:
                raise FtlMrtHeaderError(f'MRT record too large ({length}B)', data=header_bytes)

            # Update stats record
            if FTL_RECORD_BGP_STATS:

                # Add entry
                if FTL_ATTR_BGP_STATS_MRT_ENTRIES >= 0:
                    stats_record[FTL_ATTR_BGP_STATS_MRT_ENTRIES] += 1

                # Add entry type
                if FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES >= 0:
                    stats_record_mrt_entry_types = stats_record[FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES]
                    etype, estype = mrt_type, mrt_subtype
                    if FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPES_HUMAN:
                        estype = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TO_STR.get(etype, dict()).get(estype, estype)
                        etype = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPE_TO_STR.get(etype, etype)
                    etype_str = f'{etype}|{estype}'
                    stats_record_mrt_entry_types[etype_str] = stats_record_mrt_entry_types.get(etype_str, 0) + 1

            # Slice MRT data bytes to current entry length
            entry_bytes = data_bytes[:length]

            # Read MRT entry bytes
            n_read = 0
            try:
                n_read = inputfile.readinto(entry_bytes)
            except Exception:  # pylint: disable=broad-except
                pass

            # Check MRT entry bytes
            if n_read < length:
                raise FtlMrtHeaderError(f'Incomplete MRT record ({n_read}B<{length}B)', data=header_bytes)

            ##########
            # BGP4MP #
            ##########

            # Parse MRT BGP4MP entry
            if mrt_type == MRT_BGP4MP:

                # Increase sequence number
                sequence += 1

                # Yield BGP4MP records
                yield from unpack_mrt_entry_bgp4mp(caches, stats_record, bgp_error, state_change_records, route_records,
                                                   keep_alive_records, route_refresh_records, notification_records,
                                                   open_records, entry_bytes, mrt_subtype, sequence, ts)

            #############
            # BGP4MP_ET #
            #############

            # Parse MRT BGP4MP_ET entry
            elif mrt_type == MRT_BGP4MP_ET:

                # ------------------------------------------
                # [RFC6396] 3. Extended Timestamp MRT Header
                # ------------------------------------------
                #  0                   1                   2                   3
                #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |                           Timestamp                           |
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |             Type              |            Subtype            |
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |                             Length                            |
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |                      Microsecond Timestamp                    |
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |                      Message... (variable)
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

                # Increase sequence number
                sequence += 1

                # Parse millisecond timestamp
                ts += struct_unpack(STRUCT_4B, entry_bytes[:4])[0] / 1000000.0
                entry_bytes = entry_bytes[4:]

                # Yield BGP4MP records
                yield from unpack_mrt_entry_bgp4mp(caches, stats_record, bgp_error, state_change_records, route_records,
                                                   keep_alive_records, route_refresh_records, notification_records,
                                                   open_records, entry_bytes, mrt_subtype, sequence, ts)

            #################
            # TABLE_DUMP_V2 #
            #################

            # Parse MRT TABLE_DUMP_V2 entry
            elif mrt_type == MRT_TABLE_DUMP_V2:

                # Extract peer table
                if mrt_subtype == MRT_TABLE_DUMP_V2_PEER_INDEX_TABLE:
                    yield from unpack_mrt_entry_tdv2_index(caches, stats_record, peer_table_records, entry_bytes,
                                                           peer_table)

                    # Finalize peer table
                    peer_table = tuple(peer_table)
                    continue

                # ------------------------------------------------
                # [RFC8050] 4. MRT Subtypes for Type TABLE_DUMP_V2
                # ------------------------------------------------
                # This document defines the following new subtypes:
                #  o  RIB_IPV4_UNICAST_ADDPATH
                #  o  RIB_IPV4_MULTICAST_ADDPATH
                #  o  RIB_IPV6_UNICAST_ADDPATH
                #  o  RIB_IPV6_MULTICAST_ADDPATH
                #  o  RIB_GENERIC_ADDPATH

                # Check for ADD-PATH entry (RFC7911)
                addpath = mrt_subtype in MRT_TABLE_DUMP_V2_RIB_ADDPATH_ANY

                # Yield RIB records
                yield from unpack_mrt_entry_tdv2_rib(caches, stats_record, route_records, entry_bytes, mrt_subtype,
                                                     peer_table, addpath=addpath)

            ##############
            # TABLE_DUMP #
            ##############

            # Parse MRT TABLE_DUMP entry
            elif mrt_type == MRT_TABLE_DUMP:

                # Yield RIB records
                yield from unpack_mrt_entry_td_rib(caches, stats_record, route_records, entry_bytes, mrt_subtype)

            #######
            # BGP #
            #######

            # Parse MRT BGP entry
            elif mrt_type == MRT_BGP:

                # Increase sequence number
                sequence += 1

                # Yield BGP records
                yield from unpack_mrt_entry_bgp(caches, stats_record, bgp_error, state_change_records, route_records,
                                                keep_alive_records, route_refresh_records, notification_records,
                                                open_records, entry_bytes, mrt_subtype, sequence, ts)

            # Successfully parsed first record
            first_record = False

        # Re-raise already handled header exceptions as format exceptions
        except FtlMrtHeaderError as error:

            # Raise format error for invalid first header
            if first_record is True:
                raise FtlMrtFormatError(error=error)  # pylint: disable=raise-missing-from

            # Yield (or re-raise) other header errors
            # NOTE: We abort parsing in case of any header error
            yield from bgp_error(error)
            return

        # Re-raise already handled record errors
        except FtlMrtError as error:
            raise error

        # Yield (or re-raise) unhandled exceptions
        except Exception as exc:  # pylint: disable=broad-exception-caught
            yield from bgp_error(FtlMrtDataError('Unhandled data error', data=b''.join((header_bytes, entry_bytes)),
                                                 trace=True, exception=exc))



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import base64

# Local imports
from .const import BGP_SAFI_UNICAST
from .const import BGP_SAFI_MULTICAST
from .const import BGP_PATH_ATTR_ORIGIN
from .const import BGP_PATH_ATTR_AS_PATH
from .const import BGP_PATH_ATTR_AS4_PATH
from .const import BGP_PATH_ATTR_AS_PATH_SEGMENT_SET
from .const import BGP_PATH_ATTR_AS_PATH_SEGMENT_CONFED_SET
from .const import BGP_PATH_ATTR_AS4_PATH_AS_TRANS
from .const import BGP_PATH_ATTR_NEXT_HOP
from .const import BGP_PATH_ATTR_COMMUNITIES
from .const import BGP_PATH_ATTR_LARGE_COMMUNITIES
from .const import BGP_PATH_ATTR_EXTENDED_COMMUNITIES
from .const import BGP_PATH_ATTR_MULTI_EXIT_DISC
from .const import BGP_PATH_ATTR_MP_REACH_NLRI
from .const import BGP_PATH_ATTR_MP_UNREACH_NLRI
from .const import BGP_PATH_ATTR_ATOMIC_AGGREGATE
from .const import BGP_PATH_ATTR_AGGREGATOR
from .const import BGP_PATH_ATTR_AS4_AGGREGATOR
from .const import BGP_PATH_ATTR_ONLY_TO_CUSTOMER
from .const import BGP_PATH_ATTR_ORIGINATOR_ID
from .const import BGP_PATH_ATTR_CLUSTER_LIST
from .const import BGP_PATH_ATTR_LOCAL_PREF
from .const import BGP_PATH_ATTR_ATTR_SET
from .const import BGP_PATH_ATTR_AS_PATHLIMIT
from .const import BGP_PATH_ATTR_AIGP
from .nlri import unpack_mrt_bgp_nlri
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import AFI_IPV4
from ...const import AFI_IPV6
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import STRUCT_8B
from ...const import STRUCT_2B2B
from ...const import STRUCT_8B8B
from ...const import STRUCT_2B2B2B
from ...const import STRUCT_4B4B4B
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_ASPATH
from ....model.attr import FTL_ATTR_BGP_ROUTE_ORIGIN
from ....model.attr import FTL_ATTR_BGP_ROUTE_COMMUNITIES
from ....model.attr import FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES
from ....model.attr import FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES
from ....model.attr import FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC
from ....model.attr import FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE
from ....model.attr import FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS
from ....model.attr import FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER
from ....model.attr import FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID
from ....model.attr import FTL_ATTR_BGP_ROUTE_CLUSTER_LIST
from ....model.attr import FTL_ATTR_BGP_ROUTE_LOCAL_PREF
from ....model.attr import FTL_ATTR_BGP_ROUTE_ATTR_SET
from ....model.attr import FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT
from ....model.attr import FTL_ATTR_BGP_ROUTE_AIGP
from ....model.attr import FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_COMMUNITIES_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_CLUSTER_LIST_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES_HUMAN
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_IGP
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_EGP
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE
from ....model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR
from ....model.const import FTL_ATTR_BGP_STATS_BGP_PATH_ATTR_TO_STR
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH_STR
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH_STR
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtDataError


def unpack_mrt_bgp_attr(caches, stats_record, route_init, route_emit, route_record, attr_bytes, aslen=4, addpath=False,
                        rib=False):
    """ Parse MRT attributes.
    """
    # Prepare AS byte length
    asbytelen = STRUCT_2B if aslen == 2 else STRUCT_4B

    # Initialize AS/AS4 path support
    aspath, aggr_as, aggr_ip = None, None, None
    as4path, aggr4_as, aggr4_ip = None, None, None

    # Initialize announced/withdrawn prefixes
    nlri, nlui = tuple(), tuple()

    #####################
    # ATTRIBUTE PARSING #
    #####################

    # Prepare byte offset
    offset = 0

    # ------------------------------------
    # [RFC4271] 4.3. UPDATE Message Format
    # ------------------------------------
    # 0                   1
    # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |  Attr. Flags  |Attr. Type Code|
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Parse attributes
    attrlen = len(attr_bytes)
    while offset < attrlen:

        # Parse attribute flags
        flags = attr_bytes[offset]
        offset += 1

        # Parse attribute type
        atype = attr_bytes[offset]
        offset += 1

        # Update stats record
        if FTL_RECORD_BGP_STATS:

            # Add attribute type
            if FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES >= 0:
                attrtype = atype
                if FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES_HUMAN:
                    attrtype = FTL_ATTR_BGP_STATS_BGP_PATH_ATTR_TO_STR.get(atype, atype)
                attrtype = str(attrtype)
                stats_record_mrt_bgp_attr = stats_record[FTL_ATTR_BGP_STATS_MRT_BGP_ATTRIBUTE_TYPES]
                stats_record_mrt_bgp_attr[attrtype] = stats_record_mrt_bgp_attr.get(attrtype, 0) + 1

        # ------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format
        # ------------------------------------
        # The fourth high-order bit (bit 3) of the Attribute Flags octet
        # is the Extended Length bit. It defines whether the Attribute
        # Length is one octet (if set to 0) or two octets (if set to 1).
        #
        # If the Extended Length bit of the Attribute Flags octet is set
        # to 0, the third octet of the Path Attribute contains the length
        # of the attribute data in octets.
        #
        # If the Extended Length bit of the Attribute Flags octet is set
        # to 1, the third and fourth octets of the path attribute contain
        # the length of the attribute data in octets.

        # Parse attribute length
        alen = 0
        if flags & 0b00010000:
            alen = struct_unpack(STRUCT_2B, attr_bytes[offset:offset + 2])[0]
            offset += 2
        else:
            alen = attr_bytes[offset]
            offset += 1

        # Prepare attribute byte offsets (current/end)
        cur_offset, end_offset = offset, offset + alen
        offset = end_offset

        # Sanitize attribute
        if offset > attrlen:
            # NOTE: A few UPDATE messages (3-5) from RouteViews Oregon2 occasionally include incomplete BGP attributes
            raise FtlMrtDataError(f'Incomplete BGP attribute ({attrlen - cur_offset}B<{alen}B)',
                                  reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR, data=attr_bytes)

        ####################
        # ORIGIN ATTRIBUTE #
        ####################

        # --------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - a) ORIGIN (Type Code 1)
        # --------------------------------------------------------------
        # ORIGIN is a well-known mandatory attribute that defines the
        # origin of the path information. The data octet can assume
        # the following values:
        #
        #    Value      Meaning
        #
        #    0         IGP - Network Layer Reachability Information
        #                 is interior to the originating AS
        #
        #    1         EGP - Network Layer Reachability Information
        #                 learned via the EGP protocol [RFC904]
        #
        #    2         INCOMPLETE - Network Layer Reachability
        #                 Information learned by some other means

        # Parse origin attribute
        if atype == BGP_PATH_ATTR_ORIGIN:
            if FTL_ATTR_BGP_ROUTE_ORIGIN >= 0:
                origin = attr_bytes[cur_offset]
                if FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN:
                    if origin == FTL_ATTR_BGP_ROUTE_ORIGIN_IGP:
                        origin = FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR
                    elif origin == FTL_ATTR_BGP_ROUTE_ORIGIN_EGP:
                        origin = FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR
                    elif origin == FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE:
                        origin = FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR
                route_record[FTL_ATTR_BGP_ROUTE_ORIGIN] = origin

        #####################
        # AS_PATH ATTRIBUTE #
        #####################

        # ---------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - b) AS_PATH (Type Code 2)
        # ---------------------------------------------------------------
        # AS_PATH is a well-known mandatory attribute that is composed
        # of a sequence of AS path segments. Each AS path segment is
        # represented by a triple <path segment type, path segment
        # length, path segment value>.
        #
        # The path segment type is a 1-octet length field with the
        # following values defined:
        #
        #    Value      Segment Type
        #
        #    1         AS_SET: unordered set of ASes a route in the
        #                 UPDATE message has traversed
        #
        #    2         AS_SEQUENCE: ordered set of ASes a route in
        #                 the UPDATE message has traversed
        #
        # The path segment length is a 1-octet length field,
        # containing the number of ASes (not the number of octets) in
        # the path segment value field.
        #
        # The path segment value field contains one or more AS
        # numbers, each encoded as a 2-octet length field.

        # -----------------------------
        # [RFC6793] Protocol Extensions
        # -----------------------------
        # This document defines a new BGP path attribute called AS4_PATH.
        # This is an optional transitive attribute that contains the AS path
        # encoded with four-octet AS numbers. The AS4_PATH attribute has the
        # same semantics and the same encoding as the AS_PATH attribute,
        # except that it is "optional transitive", and it carries four-octet
        # AS numbers.

        # Parse AS/AS4 path attribute
        # pylint: disable-next=confusing-consecutive-elif,consider-using-in
        elif atype == BGP_PATH_ATTR_AS_PATH or atype == BGP_PATH_ATTR_AS4_PATH:
            if FTL_ATTR_BGP_ROUTE_ASPATH >= 0:

                # Sanitize AS path attribute
                # NOTE: Some MRT exporters fail to set correct AS/AS4 path type
                hoplen, hopbytelen = aslen, asbytelen
                c_offset2, c_offset4 = cur_offset, cur_offset
                do_continue = True
                while do_continue:
                    do_continue = False
                    if c_offset2 < end_offset:
                        c_offset2 += attr_bytes[c_offset2 + 1] * 2 + 2
                        do_continue = True
                    if c_offset4 < end_offset:
                        c_offset4 += attr_bytes[c_offset4 + 1] * 4 + 2
                        do_continue = True
                if c_offset2 == c_offset4 == end_offset:
                    hoplen, hopbytelen = aslen, asbytelen
                elif c_offset4 == end_offset:
                    hoplen, hopbytelen = 4, STRUCT_4B
                elif c_offset2 == end_offset:
                    hoplen, hopbytelen = 2, STRUCT_2B
                else:
                    hlen = end_offset - cur_offset
                    chlen = (c_offset4 if aslen == 4 else c_offset2) - cur_offset
                    raise FtlMrtDataError(f'Incomplete AS path attribute ({chlen}>{hlen}B)',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH, data=attr_bytes)

                # Check for AS length fixes
                if hoplen != aslen:

                    # Update stats record
                    if FTL_RECORD_BGP_STATS:

                        # Add AS length fix
                        if FTL_ATTR_BGP_STATS_MRT_FIXES >= 0:
                            fixtype = str(FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH if rib is False
                                          else FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH)
                            if FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN:
                                fixtype = (FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH_STR if rib is False
                                           else FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH_STR)
                            stats_record_fix = stats_record[FTL_ATTR_BGP_STATS_MRT_FIXES]
                            stats_record_fix[fixtype] = stats_record_fix.get(fixtype, 0) + 1

                # Parse AS path segments
                cur_aspath = list()
                while cur_offset < end_offset:

                    # Parse path segment type
                    stype = attr_bytes[cur_offset]
                    cur_offset += 1

                    # Parse path segment length
                    slen = attr_bytes[cur_offset]
                    cur_offset += 1

                    # Parse path segment or set
                    target, asset = cur_aspath, False
                    # pylint: disable-next=consider-using-in
                    if stype == BGP_PATH_ATTR_AS_PATH_SEGMENT_SET or stype == BGP_PATH_ATTR_AS_PATH_SEGMENT_CONFED_SET:
                        target, asset = list(), True

                    # Parse path segment/set value
                    for cur_offset in range(cur_offset + hoplen, cur_offset + hoplen * slen + hoplen, hoplen):
                        target.append(struct_unpack(hopbytelen, attr_bytes[cur_offset - hoplen:cur_offset])[0])

                    # Finalize path set
                    if asset is True:
                        cur_aspath.append(tuple(sorted(set(target))))

                # --------------------------------------------
                # [RFC6793] 4.2.3. Processing Received Updates
                # --------------------------------------------
                # When a NEW BGP speaker receives an update from an OLD BGP speaker, it
                # MUST be prepared to receive the AS4_PATH attribute along with the
                # existing AS_PATH attribute. If the AS4_PATH attribute is also
                # received, both of the attributes will be used to construct the exact
                # AS path information, and therefore the information carried by both of
                # the attributes will be considered for AS path loop detection.

                # Update AS path
                if atype == BGP_PATH_ATTR_AS_PATH:
                    aspath = cur_aspath
                elif atype == BGP_PATH_ATTR_AS4_PATH:
                    as4path = cur_aspath

        #########################
        # COMMUNITIES ATTRIBUTE #
        #########################

        # -------------------------------
        # [RFC1997] COMMUNITIES attribute
        # -------------------------------
        # This document creates the COMMUNITIES path attribute is an optional
        # transitive attribute of variable length. The attribute consists of a
        # set of four octet values, each of which specify a community. All
        # routes with this attribute belong to the communities listed in the
        # attribute.
        #
        # The community attribute values shall be encoded using an autonomous
        # system number in the first two octets. The semantics of the final
        # two octets may be defined by the autonomous system.

        # Parse communities attribute
        elif atype == BGP_PATH_ATTR_COMMUNITIES:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_COMMUNITIES >= 0:
                if FTL_ATTR_BGP_ROUTE_COMMUNITIES_HUMAN:
                    route_record[FTL_ATTR_BGP_ROUTE_COMMUNITIES] = tuple(
                        '{}:{}'.format(*struct_unpack(STRUCT_2B2B, attr_bytes[c_offset:c_offset + 4]))
                        for c_offset in range(cur_offset, end_offset, 4))
                else:
                    route_record[FTL_ATTR_BGP_ROUTE_COMMUNITIES] = tuple(
                        struct_unpack(STRUCT_4B, attr_bytes[c_offset:c_offset + 4])[0]
                        for c_offset in range(cur_offset, end_offset, 4))

        ######################
        # NEXT_HOP ATTRIBUTE #
        ######################

        # ----------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - c) NEXT_HOP (Type Code 3)
        # ----------------------------------------------------------------
        # This is a well-known mandatory attribute that defines the
        # (unicast) IP address of the router that SHOULD be used as
        # the next hop to the destinations listed in the Network Layer
        # Reachability Information field of the UPDATE message.

        # Parse nexthop IP attribute (IPv4 only)
        elif atype == BGP_PATH_ATTR_NEXT_HOP:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL >= 0:
                nexthop_proto = IPV4_STR if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN else IPV4
                route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL] = nexthop_proto
            if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP >= 0:
                nexthop_ip = attr_bytes[cur_offset:end_offset]
                if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN:
                    route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_IP] = socket_inet_ntop(AF_INET, nexthop_ip)
                else:
                    route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_IP] = struct_unpack(STRUCT_4B, nexthop_ip)[0]

        ###########################
        # MP_REACH_NLRI ATTRIBUTE #
        ###########################

        # ------------------------------------------------------------------------
        # [RFC4760] 3. Multiprotocol Reachable NLRI - MP_REACH_NLRI (Type Code 14)
        # ------------------------------------------------------------------------
        # This is an optional non-transitive attribute that can be used for the
        # following purposes:
        #
        # (a) to advertise a feasible route to a peer
        #
        # (b) to permit a router to advertise the Network Layer address of the
        #     router that should be used as the next hop to the destinations
        #     listed in the Network Layer Reachability Information field of the
        #     MP_NLRI attribute.
        #
        # +---------------------------------------------------------+
        # | Address Family Identifier (2 octets)                    |
        # +---------------------------------------------------------+
        # | Subsequent Address Family Identifier (1 octet)          |
        # +---------------------------------------------------------+
        # | Length of Next Hop Network Address (1 octet)            |
        # +---------------------------------------------------------+
        # | Network Address of Next Hop (variable)                  |
        # +---------------------------------------------------------+
        # | Reserved (1 octet)                                      |
        # +---------------------------------------------------------+
        # | Network Layer Reachability Information (variable)       |
        # +---------------------------------------------------------+
        #
        # An UPDATE message that carries the MP_REACH_NLRI MUST also carry the
        # ORIGIN and the AS_PATH attributes (both in EBGP and in IBGP
        # exchanges). Moreover, in IBGP exchanges such a message MUST also
        # carry the LOCAL_PREF attribute.
        #
        # An UPDATE message that carries no NLRI, other than the one encoded in
        # the MP_REACH_NLRI attribute, SHOULD NOT carry the NEXT_HOP attribute.
        # If such a message contains the NEXT_HOP attribute, the BGP speaker
        # that receives the message SHOULD ignore this attribute.

        # Parse reachable prefix attribute
        elif atype == BGP_PATH_ATTR_MP_REACH_NLRI:  # pylint: disable=confusing-consecutive-elif
            afinet = None

            # Parse UPDATE message (including AFI/SAFI)
            if rib is False:

                # Parse AFI value
                afi = struct_unpack(STRUCT_2B, attr_bytes[cur_offset:cur_offset + 2])[0]
                cur_offset += 2

                # Parse SAFI value
                safi = attr_bytes[cur_offset]
                cur_offset += 1

                # Check AFI value
                if afi != AFI_IPV4 and afi != AFI_IPV6:  # pylint: disable=consider-using-in
                    raise FtlMrtDataError(f'Invalid AFI value ({afi}) in MP_REACH_NLRI attribute',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=attr_bytes)

                # Check SAFI value
                if safi != BGP_SAFI_UNICAST and safi != BGP_SAFI_MULTICAST:  # pylint: disable=consider-using-in
                    raise FtlMrtDataError(f'Unsupported SAFI value ({safi}) in MP_REACH_NLRI attribute',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=attr_bytes)

                # Parse protocol
                afinet = AF_INET6 if afi == AFI_IPV6 else AF_INET

            # ----------------------------
            # [RFC6396] 4.3.4. RIB Entries
            # ----------------------------
            # There is one exception to the encoding of BGP attributes for the BGP
            # MP_REACH_NLRI attribute (BGP Type Code 14) [RFC4760]. Since the AFI,
            # SAFI, and NLRI information is already encoded in the RIB Entry Header
            # or RIB_GENERIC Entry Header, only the Next Hop Address Length and
            # Next Hop Address fields are included. The Reserved field is omitted.
            # The attribute length is also adjusted to reflect only the length of
            # the Next Hop Address Length and Next Hop Address fields.

            # ----------------
            # [RFC6396] Errata
            # ----------------
            # The encoding of the MP_REACH_NLRI attribute is not in the form
            # according to Section 4.3.4. RIB Entries.
            #
            # NOTE: This seems to be an IPv6-related issue
            #
            # The example includes a full MP_REACH_NLRI attribute. This is a common
            # issue with TABLE_DUMP_V2 and parsers need to implement a workaround to
            # support the broken form.
            #
            # One way of solving this is to compare the attribute length of
            # MP_REACH_NLRI with the first byte of the attribute. If the value of the
            # first byte is equal to the attribute lenght - 1 then it is the RFC
            # encoding else assume that a full MP_REACH_NLRI attribute was dumped in
            # which case the parser needs to skip the first 3 bytes to get to the
            # nexthop.

            # Parse RIB entry (with or without AFI/SAFI)
            elif attr_bytes[cur_offset] != alen - 1:
                cur_offset += 3

                # Update stats record
                if FTL_RECORD_BGP_STATS:

                    # Add ADD-PATH fix
                    if FTL_ATTR_BGP_STATS_MRT_FIXES >= 0:
                        fixtype = str(FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI)
                        if FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN:
                            fixtype = FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI_STR
                        stats_record_fix = stats_record[FTL_ATTR_BGP_STATS_MRT_FIXES]
                        stats_record_fix[fixtype] = stats_record_fix.get(fixtype, 0) + 1

            # Parse nexthop length
            nhlen = attr_bytes[cur_offset]
            cur_offset += 1

            # --------------------------------------------
            # [RFC2545] 3. Constructing the Next Hop field
            # --------------------------------------------
            # A BGP speaker shall advertise to its peer in the Network Address of
            # Next Hop field the global IPv6 address of the next hop, potentially
            # followed by the link-local IPv6 address of the next hop.
            #
            # NOTE: We use global IPv6 addresses only, but skip link-local addresses
            #
            # The value of the Length of Next Hop Network Address field on a
            # MP_REACH_NLRI attribute shall be set to 16, when only a global
            # address is present, or 32 if a link-local address is also included in
            # the Next Hop field.

            # Skip link-local address
            nhiplen = nhlen if nhlen != 32 else 16
            mp_afinet = AF_INET6 if nhiplen == 16 else AF_INET

            # Parse nexthop protocol
            mp_nexthop_proto = None
            if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL >= 0:
                if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN:
                    mp_nexthop_proto = IPV6_STR if mp_afinet == AF_INET6 else IPV4_STR
                else:
                    mp_nexthop_proto = IPV6 if mp_afinet == AF_INET6 else IPV4
                if rib is True:
                    route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL] = mp_nexthop_proto

            # Parse nexthop IP
            mp_nexthop_ip = None
            if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP >= 0:
                nhiplen = nhlen if nhlen != 32 else 16
                mp_afinet = AF_INET6 if nhiplen == 16 else AF_INET
                mp_nexthop_ip = attr_bytes[cur_offset:cur_offset + nhiplen]
                if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN:
                    mp_nexthop_ip = socket_inet_ntop(mp_afinet, mp_nexthop_ip)
                elif mp_afinet == AF_INET6:
                    net, host = struct_unpack(STRUCT_8B8B, mp_nexthop_ip)
                    mp_nexthop_ip = (net << 64) + host
                else:
                    mp_nexthop_ip = struct_unpack(STRUCT_4B, mp_nexthop_ip)[0]
                # NOTE: If there is an MP NEXT_HOP for RIB entries, we overwrite any existing pre-MP NEXT_HOP attribute,
                # NOTE: since there is only one prefix per attribute set (provided outside of MP_REACH_NLRI attributes)
                if rib is True:
                    route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_IP] = mp_nexthop_ip
            cur_offset += nhlen

            # Parse prefix
            if afinet is not None:

                # Skip reserved byte
                cur_offset += 1

                # Parse NLRI
                # NOTE: For UPDATE messages, we do not overwrite the pre-MP NEXT_HOP attribute, but hand over the MP
                # NOTE: NEXT_HOP to unpack_mrt_bgp_nlri() instead.
                if cur_offset < end_offset:
                    nlri += unpack_mrt_bgp_nlri(caches, stats_record, attr_bytes[cur_offset:end_offset], afinet,
                                                nexthop_proto=mp_nexthop_proto, nexthop_ip=mp_nexthop_ip,
                                                addpath=addpath)

        #############################
        # MULTI_EXIT_DISC ATTRIBUTE #
        #############################

        # -----------------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - d) MULTI_EXIT_DISC (Type Code 4)
        # -----------------------------------------------------------------------
        # This is an optional non-transitive attribute that is a
        # four-octet unsigned integer. The value of this attribute
        # MAY be used by a BGP speaker's Decision Process to
        # discriminate among multiple entry points to a neighboring
        # autonomous system.

        # Parse MED metric attribute
        elif atype == BGP_PATH_ATTR_MULTI_EXIT_DISC:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC] = struct_unpack(STRUCT_4B,
                                                                                 attr_bytes[cur_offset:end_offset])[0]

        ###############################
        # LARGE COMMUNITIES ATTRIBUTE #
        ###############################

        # --------------------------------------------
        # [RFC8092} 3. BGP Large Communities Attribute
        # --------------------------------------------
        # This document defines the BGP Large Communities attribute as an
        # optional transitive path attribute of variable length. All routes
        # with the BGP Large Communities attribute belong to the communities
        # specified in the attribute.
        #
        # Each BGP Large Community value is encoded as a 12-octet quantity, as
        # follows:
        #
        #  0                   1                   2                   3
        #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                      Global Administrator                     |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                       Local Data Part 1                       |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                       Local Data Part 2                       |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        # Parse large communities attribute
        elif atype == BGP_PATH_ATTR_LARGE_COMMUNITIES:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES >= 0:
                if FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES_HUMAN:
                    route_record[FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES] = tuple(
                        '{}:{}:{}'.format(*struct_unpack(STRUCT_4B4B4B, attr_bytes[c_offset:c_offset + 12]))
                        for c_offset in range(cur_offset, end_offset, 12))
                else:
                    route_record[FTL_ATTR_BGP_ROUTE_LARGE_COMMUNITIES] = tuple(
                        ((struct_unpack(STRUCT_8B, attr_bytes[c_offset:c_offset + 8])[0] << 32)
                         + struct_unpack(STRUCT_4B, attr_bytes[c_offset + 8:c_offset + 12])[0])
                        for c_offset in range(cur_offset, end_offset, 12))

        #############################
        # MP_UNREACH_NLRI ATTRIBUTE #
        #############################

        # ----------------------------------------------------------------------------
        # [RFC4760] 4. Multiprotocol Unreachable NLRI - MP_UNREACH_NLRI (Type Code 15)
        # ----------------------------------------------------------------------------
        # This is an optional non-transitive attribute that can be used for the
        # purpose of withdrawing multiple unfeasible routes from service.
        #
        # +---------------------------------------------------------+
        # | Address Family Identifier (2 octets)                    |
        # +---------------------------------------------------------+
        # | Subsequent Address Family Identifier (1 octet)          |
        # +---------------------------------------------------------+
        # | Withdrawn Routes (variable)                             |
        # +---------------------------------------------------------+
        #
        # An UPDATE message that contains the MP_UNREACH_NLRI is not required
        # to carry any other path attributes.

        # Parse unreachable prefix attribute
        elif atype == BGP_PATH_ATTR_MP_UNREACH_NLRI:  # pylint: disable=confusing-consecutive-elif

            # Parse UPDATE message
            if rib is False:

                # Parse AFI value
                afi = struct_unpack(STRUCT_2B, attr_bytes[cur_offset:cur_offset + 2])[0]
                cur_offset += 2

                # Parse SAFI value
                safi = attr_bytes[cur_offset]
                cur_offset += 1

                # Check AFI value
                if afi != AFI_IPV4 and afi != AFI_IPV6:  # pylint: disable=consider-using-in
                    raise FtlMrtDataError(f'Invalid AFI value ({afi}) in MP_UNREACH_NLRI attribute',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=attr_bytes)

                # Check SAFI value
                if safi != BGP_SAFI_UNICAST and safi != BGP_SAFI_MULTICAST:  # pylint: disable=consider-using-in
                    raise FtlMrtDataError(f'Unsupported SAFI value ({safi}) in MP_UNREACH_NLRI attribute',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=attr_bytes)

                # Parse protocol
                afinet = AF_INET6 if afi == AFI_IPV6 else AF_INET

                # Parse NLRI
                if cur_offset < end_offset:
                    nlui += unpack_mrt_bgp_nlri(caches, stats_record, attr_bytes[cur_offset:end_offset], afinet,
                                                addpath=addpath)

        ########################
        # AGGREGATOR ATTRIBUTE #
        ########################

        # ------------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - g) AGGREGATOR (Type Code 7)
        # ------------------------------------------------------------------
        # AGGREGATOR is an optional transitive attribute of length 6.
        # The attribute contains the last AS number that formed the
        # aggregate route (encoded as 2 octets), followed by the IP
        # address of the BGP speaker that formed the aggregate route
        # (encoded as 4 octets). This SHOULD be the same address as
        # the one used for the BGP Identifier of the speaker.

        # -----------------------------
        # [RFC6793] Protocol Extensions
        # -----------------------------
        # This document defines a new BGP path attribute called
        # AS4_AGGREGATOR, which is optional transitive. The AS4_AGGREGATOR
        # attribute has the same semantics and the same encoding as the
        # AGGREGATOR attribute, except that it carries a four-octet AS number.

        # Parse AS/AS4 aggregator attribute (IPv4 only)
        # pylint: disable-next=consider-using-in,confusing-consecutive-elif
        elif atype == BGP_PATH_ATTR_AGGREGATOR or atype == BGP_PATH_ATTR_AS4_AGGREGATOR:
            if (FTL_ATTR_BGP_ROUTE_ASPATH >= 0 or FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL >= 0
                or FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS >= 0 or FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP >= 0):

                # Sanitize aggregator
                # NOTE: Some MRT exporters fail to set correct AS/AS4 aggregator type
                aggasbytelen = STRUCT_2B
                if end_offset - cur_offset == 8:
                    aggasbytelen = STRUCT_4B
                elif end_offset - cur_offset != 6:
                    cagglen = end_offset - cur_offset
                    agglen = 8 if atype == BGP_PATH_ATTR_AS4_AGGREGATOR else 6
                    raise FtlMrtDataError(f'Incomplete aggregator attribute ({cagglen}!={agglen}B)',
                                          reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR, data=attr_bytes)

                # Check for AS length fixes
                if aggasbytelen != asbytelen:

                    # Update stats record
                    if FTL_RECORD_BGP_STATS:

                        # Add AS length fix
                        if FTL_ATTR_BGP_STATS_MRT_FIXES >= 0:
                            fixtype = str(FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH if rib is False
                                          else FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH)
                            if FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN:
                                fixtype = (FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH_STR if rib is False
                                           else FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH_STR)
                            stats_record_fix = stats_record[FTL_ATTR_BGP_STATS_MRT_FIXES]
                            stats_record_fix[fixtype] = stats_record_fix.get(fixtype, 0) + 1

                # Parse aggregator
                aggas = struct_unpack(aggasbytelen, attr_bytes[cur_offset:end_offset - 4])[0]
                aggip = attr_bytes[end_offset - 4:end_offset]
                if atype == BGP_PATH_ATTR_AGGREGATOR:
                    aggr_as, aggr_ip = aggas, aggip
                else:
                    aggr4_as, aggr4_ip = aggas, aggip

        ##################################
        # EXTENDED COMMUNITIES ATTRIBUTE #
        ##################################

        # -----------------------------------------------
        # [RFC4360] 2. BGP Extended Communities Attribute
        # -----------------------------------------------
        # The Extended Communities Attribute is a transitive optional BGP
        # attribute, with the Type Code 16. The attribute consists of a set of
        # "extended communities". All routes with the Extended Communities
        # attribute belong to the communities listed in the attribute.
        #
        # Each Extended Community is encoded as an 8-octet quantity, as
        # follows:
        #
        #    - Type Field  : 1 or 2 octets
        #    - Value Field : Remaining octets
        #
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |  Type high    |  Type low(*)  |                               |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+          Value                |
        #    |                                                               |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #
        #    (*) Present for Extended types only, used for the Value field
        #        otherwise.

        # Parse extended communities attribute
        elif atype == BGP_PATH_ATTR_EXTENDED_COMMUNITIES:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES >= 0:
                if FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES_HUMAN:
                    route_record[FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES] = tuple(
                        '0x{:02x}:0x{:02x}:{}:{}:{}'.format(
                            attr_bytes[c_offset], attr_bytes[c_offset + 1],
                            *struct_unpack(STRUCT_2B2B2B, attr_bytes[c_offset + 2:c_offset + 8])
                        ) for c_offset in range(cur_offset, end_offset, 8)
                    )
                else:
                    route_record[FTL_ATTR_BGP_ROUTE_EXTENDED_COMMUNITIES] = tuple(
                        struct_unpack(STRUCT_8B, attr_bytes[c_offset:c_offset + 8])[0]
                        for c_offset in range(cur_offset, end_offset, 8)
                    )

        ##############################
        # ATOMIC_AGGREGATE ATTRIBUTE #
        ##############################

        # ------------------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - f) ATOMIC_AGGREGATE (Type Code 6)
        # ------------------------------------------------------------------------
        # ATOMIC_AGGREGATE is a well-known discretionary attribute of
        # length 0.

        # Parse atomic aggregate attribute
        elif atype == BGP_PATH_ATTR_ATOMIC_AGGREGATE:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_ATOMIC_AGGREGATE] = True

        ##############################
        # ONLY_TO_CUSTOMER ATTRIBUTE #
        ##############################

        # ---------------------------------------------------------------------------
        # [DRAFT-IETF-IDR-BGP-OPEN-POLICY] 7. BGP Internal Only To Customer attribute
        # ---------------------------------------------------------------------------
        # The Internal Only To Customer (iOTC) attribute is a new optional,
        # non-transitive BGP Path attribute with the Type Code <TBD3>. This
        # attribute has zero length as it is used only as a flag.
        #
        # There are three rules of iOTC attribute usage:
        #
        # 1.  The iOTC attribute MUST be added to all incoming routes if the
        #     receiver's Role is Customer or Peer;
        #
        # 2.  Routes with the iOTC attribute set MUST NOT be announced by a
        #     sender whose Role is Customer or Peer;
        #
        # 3.  A sender MUST NOT include this attribute in UPDATE messages if
        #     its Role is Customer, Provider or Peer. If it is contained in an
        #     UPDATE message from eBGP speaker and receiver's Role is Customer,
        #     Provider, Peer or unspecified, then this attribute MUST be
        #     removed.
        #
        # These three rules provide mechanism that strongly prevents route leak
        # creation by an AS.

        # Parse only-to-customer attribute
        elif atype == BGP_PATH_ATTR_ONLY_TO_CUSTOMER:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_ONLY_TO_CUSTOMER] = True

        ###########################
        # ORIGINATOR_ID ATTRIBUTE #
        ###########################

        # -----------------------------------------------
        # [RFC4456] 8. Avoiding Routing Information Loops
        # -----------------------------------------------
        # ORIGINATOR_ID is a new optional, non-transitive BGP attribute of Type
        # code 9. This attribute is 4 bytes long and it will be created by an
        # RR in reflecting a route. This attribute will carry the BGP
        # Identifier of the originator of the route in the local AS. A BGP
        # speaker SHOULD NOT create an ORIGINATOR_ID attribute if one already
        # exists. A router that recognizes the ORIGINATOR_ID attribute SHOULD
        # ignore a route received with its BGP Identifier as the ORIGINATOR_ID.

        # Parse originator ID attribute (IPv4 only)
        elif atype == BGP_PATH_ATTR_ORIGINATOR_ID:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID >= 0:
                orig_bgp_id = attr_bytes[cur_offset:end_offset]
                if FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID_HUMAN:
                    route_record[FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID] = socket_inet_ntop(AF_INET, orig_bgp_id)
                else:
                    route_record[FTL_ATTR_BGP_ROUTE_ORIGINATOR_ID] = struct_unpack(STRUCT_4B, orig_bgp_id)[0]

        ##########################
        # CLUSTER_LIST ATTRIBUTE #
        ##########################

        # --------------------------
        # [RFC4456] 7. Redundant RRs
        # --------------------------
        # Usually, a cluster of clients will have a single RR. In that case,
        # the cluster will be identified by the BGP Identifier of the RR.
        # However, this represents a single point of failure so to make it
        # possible to have multiple RRs in the same cluster, all RRs in the
        # same cluster can be configured with a 4-byte CLUSTER_ID so that an RR
        # can discard routes from other RRs in the same cluster.

        # -----------------------------------------------
        # [RFC4456] 8. Avoiding Routing Information Loops
        # -----------------------------------------------
        # CLUSTER_LIST is a new, optional, non-transitive BGP attribute of Type
        # code 10. It is a sequence of CLUSTER_ID values representing the
        # reflection path that the route has passed.

        # Parse cluster list attribute (IPv4 only)
        elif atype == BGP_PATH_ATTR_CLUSTER_LIST:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_CLUSTER_LIST >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_CLUSTER_LIST] = tuple(
                    socket_inet_ntop(AF_INET, attr_bytes[c_offset:c_offset + 4])
                    if FTL_ATTR_BGP_ROUTE_CLUSTER_LIST_HUMAN
                    else struct_unpack(STRUCT_4B, attr_bytes[c_offset:c_offset + 4])[0]
                    for c_offset in range(cur_offset, end_offset, 4))

        ########################
        # LOCAL_PREF ATTRIBUTE #
        ########################

        # ------------------------------------------------------------------
        # [RFC4271] 4.3. UPDATE Message Format - e) LOCAL_PREF (Type Code 5)
        # ------------------------------------------------------------------
        # LOCAL_PREF is a well-known attribute that is a four-octet
        # unsigned integer.  A BGP speaker uses it to inform its other
        # internal peers of the advertising speaker's degree of
        # preference for an advertised route.

        # Parse local-pref attribute
        elif atype == BGP_PATH_ATTR_LOCAL_PREF:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_LOCAL_PREF >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_LOCAL_PREF] = struct_unpack(STRUCT_4B,
                                                                            attr_bytes[cur_offset:end_offset])[0]

        ##########################
        # AS_PATHLIMIT ATTRIBUTE #
        ##########################

        # -----------------------------------------------------------
        # [DRAFT-IETF-IDR-AS-PATHLIMIT] 5. The AS_PATHLIMIT Attribute
        # -----------------------------------------------------------
        # The AS_PATHLIMIT attribute is a transitive optional BGP path
        # attribute, with Type Code 21. The AS_PATHLIMIT attribute has a fixed
        # length of 5 octets. The first octet is an unsigned number that is
        # the upper bound on the number of ASes in the AS_PATH attribute of the
        # associated paths. One octet suffices because the TTL field of the IP
        # header ensures that only one octet's worth of ASes can ever be
        # traversed. The second thru fifth octets are the AS number of the AS
        # that attached the AS_PATHLIMIT attribute to the NLRI.

        # Parse AS path limit attribute
        elif atype == BGP_PATH_ATTR_AS_PATHLIMIT:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT >= 0:
                pathlimit = attr_bytes[cur_offset]
                pathlimit_origin = struct_unpack(STRUCT_4B, attr_bytes[cur_offset + 1:cur_offset + 5])[0]
                route_record[FTL_ATTR_BGP_ROUTE_AS_PATHLIMIT] = (pathlimit_origin, pathlimit)

        ######################
        # ATTR_SET ATTRIBUTE #
        ######################

        # ------------------------------------------
        # [RFC6368] 5. BGP Customer Route Attributes
        # ------------------------------------------
        # ATTR_SET is an optional transitive attribute that carries a set of
        # BGP path attributes. An attribute set (ATTR_SET) can include any
        # BGP attribute that can occur in a BGP UPDATE message, except for
        # the MP_REACH and MP_UNREACH attributes.
        #
        # The ATTR_SET attribute is encoded as follows:
        #
        # +------------------------------+
        # | Attr Flags (O|T) Code = 128  |
        # +------------------------------+
        # | Attr. Length (1 or 2 octets) |
        # +------------------------------+
        # | Origin AS (4 octets)         |
        # +------------------------------+
        # | Path Attributes (variable)   |
        # +------------------------------+
        #
        # The Attribute Flags are encoded according to RFC 4271 [RFC4271]. The
        # Extended Length bit determines whether the Attribute Length is one or
        # two octets.

        # Parse attribute-set attribute
        elif atype == BGP_PATH_ATTR_ATTR_SET:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_ATTR_SET >= 0:
                route_record_attrset = list(route_init)
                attrset_origin = struct_unpack(STRUCT_4B, attr_bytes[cur_offset:cur_offset + 4])[0]
                unpack_mrt_bgp_attr(caches, stats_record, route_init, route_emit, route_record_attrset,
                                    attr_bytes[cur_offset + 4:end_offset], aslen=aslen)
                route_record[FTL_ATTR_BGP_ROUTE_ATTR_SET] = (attrset_origin, route_emit(route_record_attrset))

        ##################
        # AIGP ATTRIBUTE #
        ##################

        # ---------------------------
        # [RFC7311] 3. AIGP Attribute
        # ---------------------------
        # The AIGP attribute is an optional, non-transitive BGP path attribute.
        # The attribute type code for the AIGP attribute is 26.
        #
        # The value field of the AIGP attribute is defined here to be a set of
        # elements encoded as "Type/Length/Value" (i.e., a set of TLVs). Each
        # such TLV is encoded as shown in Figure 1.
        #
        #  0                   1                   2                   3
        #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #  |     Type      |         Length                |               |
        #  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+               |
        #  ~                                                               ~
        #  |                           Value                               |
        #  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+..........................
        #
        #                         Figure 1: AIGP TLV
        #
        # -  Type: A single octet encoding the TLV Type. Only type 1, "AIGP
        #    TLV", is defined in this document. Use of other TLV types is
        #    outside the scope of this document.
        #
        # -  Length: Two octets encoding the length in octets of the TLV,
        #    including the Type and Length fields. The length is encoded as an
        #    unsigned binary integer. (Note that the minimum length is 3,
        #    indicating that no Value field is present.)
        #
        # -  Value: A field containing zero or more octets.
        #
        # This document defines only a single such TLV, the "AIGP TLV". The
        # AIGP TLV is encoded as follows:
        #
        # -  Type: 1
        #
        # -  Length: 11
        #
        # -  Value: Accumulated IGP Metric.
        #
        #    The value field of the AIGP TLV is always 8 octets long, and its
        #    value is interpreted as an unsigned 64-bit integer. IGP metrics
        #    are frequently expressed as 4-octet values. By using an 8-octet
        #    field, we ensure that the AIGP attribute can be used to hold the
        #    sum of an arbitrary number of 4-octet values.

        # Parse aigp attribute
        elif atype == BGP_PATH_ATTR_AIGP:  # pylint: disable=confusing-consecutive-elif
            if FTL_ATTR_BGP_ROUTE_AIGP >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_AIGP] = tuple(
                    (attr_bytes[c_offset], struct_unpack(STRUCT_8B, attr_bytes[c_offset + 3: c_offset + 11])[0])
                    for c_offset in range(cur_offset, end_offset, 11))

        # Parse unsupported attributes
        elif FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN >= 0:  # pylint: disable=confusing-consecutive-elif
            aflags, adata = bin(flags) if FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN_HUMAN else flags, None
            if cur_offset < end_offset:
                adata = attr_bytes[cur_offset:end_offset]
                if FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN_HUMAN:
                    adata = ' '.join('{:02x}'.format(byte) for byte in adata)
                else:
                    adata = base64.b64encode(adata).decode('ascii')
            if route_record[FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN] is None:
                route_record[FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN] = tuple([(aflags, atype, adata)])
            else:
                route_record[FTL_ATTR_BGP_ROUTE_ATTRS_UNKNOWN] += tuple([(aflags, atype, adata)])

    #############################
    # ATTRIBUTE POST-PROCESSING #
    #############################

    # --------------------------------------------
    # [RFC6793] 4.2.3. Processing Received Updates
    # --------------------------------------------
    # A NEW BGP speaker MUST also be prepared to receive the AS4_AGGREGATOR
    # attribute along with the AGGREGATOR attribute from an OLD BGP
    # speaker. When both of the attributes are received, if the AS number
    # in the AGGREGATOR attribute is not AS_TRANS, then:
    #
    #    -  the AS4_AGGREGATOR attribute and the AS4_PATH attribute SHALL
    #       be ignored,
    #
    #    -  the AGGREGATOR attribute SHALL be taken as the information
    #       about the aggregating node, and
    #
    #    -  the AS_PATH attribute SHALL be taken as the AS path
    #       information.
    #
    # Otherwise,
    #
    #    -  the AGGREGATOR attribute SHALL be ignored,
    #
    #    -  the AS4_AGGREGATOR attribute SHALL be taken as the information
    #       about the aggregating node, and
    #
    #    -  the AS path information would need to be constructed, as in all
    #       other cases.

    # Merge AS/AS4 aggregator attributes
    if aggr_as is not None and aggr4_as is not None:
        if aggr_as != BGP_PATH_ATTR_AS4_PATH_AS_TRANS:
            aggr4_as, aggr4_ip, as4path = None, None, None
        else:
            aggr_as, aggr_ip = aggr4_as, aggr4_ip

    # Add final aggregator attributes
    if FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS >= 0 and aggr_as is not None:
        route_record[FTL_ATTR_BGP_ROUTE_AGGREGATOR_AS] = aggr_as
    if aggr_ip is not None:
        if FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL >= 0:
            aggr_proto = IPV4_STR if FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL_HUMAN else IPV4
            route_record[FTL_ATTR_BGP_ROUTE_AGGREGATOR_PROTOCOL] = aggr_proto
        if FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP >= 0:
            if FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP_HUMAN:
                route_record[FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP] = socket_inet_ntop(AF_INET, aggr_ip)
            else:
                route_record[FTL_ATTR_BGP_ROUTE_AGGREGATOR_IP] = struct_unpack(STRUCT_4B, aggr_ip)[0]

    # --------------------------------------------
    # [RFC6793] 4.2.3. Processing Received Updates
    # --------------------------------------------
    # In order to construct the AS path information, it is necessary to
    # first calculate the number of AS numbers in the AS_PATH and AS4_PATH
    # attributes using the method specified in Section 9.1.2.2 of [RFC4271]
    # and in [RFC5065] for route selection.
    #
    # If the number of AS numbers in the AS_PATH attribute is less than the
    # number of AS numbers in the AS4_PATH attribute, then the AS4_PATH
    # attribute SHALL be ignored, and the AS_PATH attribute SHALL be taken
    # as the AS path information.
    #
    # If the number of AS numbers in the AS_PATH attribute is larger than
    # or equal to the number of AS numbers in the AS4_PATH attribute, then
    # the AS path information SHALL be constructed by taking as many AS
    # numbers and path segments as necessary from the leading part of the
    # AS_PATH attribute, and then prepending them to the AS4_PATH attribute
    # so that the AS path information has a number of AS numbers identical
    # to that of the AS_PATH attribute.
    #
    # NOTE: The following procedure is not implemented (would need a lot of state)
    #
    # Note that a valid AS_CONFED_SEQUENCE or AS_CONFED_SET path segment
    # SHALL be prepended if it is either the leading path segment or is
    # adjacent to a path segment that is prepended.

    # Merge AS/AS4 path attributes
    if FTL_ATTR_BGP_ROUTE_ASPATH >= 0:
        if aspath is not None:
            if as4path is not None:
                len_aspath, len_as4path = len(aspath), len(as4path)
                if len_aspath >= len_as4path:
                    aspath = aspath[:len_aspath - len_as4path] + as4path
            aspath = tuple(aspath)  # pylint: disable=redefined-variable-type

        # Add final AS path attribute
        # NOTE: RIB entries and announcements may include empty tuples, but should not include None (only for withdraws)
        # NOTE: -> this is NOT guaranteed, though (depends on correct usage of BGP attributes)
        route_record[FTL_ATTR_BGP_ROUTE_ASPATH] = aspath

    # Return announced/withdrawn prefixes
    return nlri, nlui
